"""
Custom widget.

This module provides customuized widgets.
"""

from PySide6.QtWidgets import (
    QWidget,
    QLabel,
    QGridLayout,
    QSizePolicy,
    QLineEdit,
    QFrame,
    QPushButton,
    QComboBox,
    QSpinBox,
    QDoubleSpinBox,
    QCheckBox,
    QSpacerItem,
    QApplication,
)
from PySide6.QtGui import QFont, Qt, QPixmap, QColor, QIcon
from PySide6.QtCore import Signal
from gcoreutils.color_palette import all_colors
from qtdraw.util.color_selector_util import _color2pixmap
from qtdraw.widget.validator import (
    validator_int,
    validator_float,
    validator_sympy_float,
    validator_sympy,
    validator_ilist,
    validator_list,
    validator_site,
    validator_bond,
    validator_site_bond,
    validator_vector_site_bond,
    validator_orbital_site_bond,
)
from qtdraw.util.latex_to_png import latex_to_png
from qtdraw.util.color_selector_util import color2pixmap


# ==================================================
class Layout(QGridLayout):
    # ==================================================
    def __init__(self, parent=None):
        """
        Layout widget.

        Args:
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setHorizontalSpacing(2)
        self.setVerticalSpacing(5)


# ==================================================
class Panel(QWidget):
    # ==================================================
    def __init__(self, parent=None):
        """
        Panel widget.

        Args:
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)
        self.layout = Layout(self)


# ==================================================
class Color(QColor):
    # ==================================================
    def __init__(self, color):
        """
        Color

        Args:
            color (str): color name.
        """
        super().__init__(all_colors[color][0])


# ==================================================
class Label(QLabel):
    # ==================================================
    def __init__(self, parent=None, text="", bold=False, color="black", size=10, math=False, dpi=120):
        """
        Label widget.

        Args:
            parent (QWidget, optional): parent.
            text (str, optional): text.
            bold (bool, optional): bold font ?
            color (str, optional): font color.
            size (int, optional): font size.
            math (bool, optional): for math ?
            dpi (int, optional): DPI for math.

        Note:
            - in math mode, text is given in LaTeX code without $.
        """
        super().__init__(parent=parent)
        policy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        self.setSizePolicy(policy)
        self.setContentsMargins(0, 0, 0, 0)
        font = QFont()
        font.setBold(bold)
        self.setFont(font)
        self.setFocusPolicy(Qt.NoFocus)
        self.setIndent(6)

        if math:
            self.to_png = lambda text: latex_to_png(text, True, color, size, dpi)
        else:
            self.to_png = None

        self.setText(text)

    # ==================================================
    def setText(self, text):
        """
        Set text.

        Args:
            text (xtr): text.
        """
        if self.to_png is None:
            super().setText(text)
        else:
            s = self.to_png(text)
            if s is None:
                return
            else:
                pixmap = QPixmap()
                pixmap.setDevicePixelRatio(1.0)
                pixmap.loadFromData(s)
                super().setPixmap(pixmap)


# ==================================================
class ColorLabel(Panel):
    # ==================================================
    def __init__(self, parent=None, color="", bold=False):
        """
        Color label widget.

        Args:
            parent (QWidget, optional): parent.
            color (str, optional): color/colormap name.
            bold (bool, optional): bold font ?
        """
        super().__init__(parent)

        label = Label(self, color, bold)
        size = label.font().pointSize()

        colorbox = _color2pixmap(color, "color_both", size)
        icon = Label(self)
        icon.setPixmap(colorbox)
        icon.setFrameStyle(QFrame.Box)
        icon.setFixedSize(colorbox.size())

        self.layout.addWidget(icon, 0, 0)
        self.layout.addWidget(label, 0, 1)


# ==================================================
class LineEdit(QLineEdit):
    _valid_style = "padding-left: 3px; background: none;"
    _invalid_style = "padding-left: 3px; background: pink;"
    _read_only_style = "padding-left: 3px; background: lightgray;"

    focusOut = Signal()

    # ==================================================
    def __init__(self, parent=None, text="", validator=None):
        """
        Line editor.

        Args:
            parent (QWidget, optional): parent.
            text (str, optional): text.
            validator (tuple, optional): (validator_type, option).
        """
        super().__init__(text, parent)
        policy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.setSizePolicy(policy)
        self.setContentsMargins(0, 0, 0, 0)

        self.set_validator(validator)
        self.set_read_only(False)
        self._raw = text
        self._backup = text
        self._valid = True

        self.setText(text)

    # ==================================================
    @property
    def validator(self):
        """
        Validator type.

        Returns:
            - (str) -- validator type.
        """
        return self._validator_type

    # ==================================================
    @property
    def read_only(self):
        """
        Read only ?

        Returns:
            - (bool) -- read only ?
        """
        return self._read_only

    # ==================================================
    def close_edit(self):
        """
        Close editor.
        """
        if self._valid:
            self.returnPressed.emit()
            self.clearFocus()

    # ==================================================
    def raw_text(self):
        """
        Raw text.

        Returns:
            - (str) -- raw text.
        """
        return self._raw

    # ==================================================
    def setText(self, text):
        """
        Set text.

        Args:
            text (str): text.
        """
        if self.read_only:
            super().setText(text)
            return

        if self._validator is None:
            s = text
        else:
            s = self._validator(text)

        if s is None:
            self.setStyleSheet(self._invalid_style)
            self._valid = False
        else:
            self.setStyleSheet(self._valid_style)
            self._backup = text
            self._raw = text
            self._valid = True
            super().setText(s)

    # ==================================================
    def set_validator(self, validator):
        """
        Set validator.

        Args:
            validator (tuple): (validator_type, option).
        """
        validator_dict = {
            "int": validator_int,
            "float": validator_float,
            "sympy_float": validator_sympy_float,
            "sympy": validator_sympy,
            "ilist": validator_ilist,
            "list": validator_list,
            "site": validator_site,
            "bond": validator_bond,
            "site_bond": validator_site_bond,
            "vector_site_bond": validator_vector_site_bond,
            "orbital_site_bond": validator_orbital_site_bond,
        }
        if validator is None:
            self._validator = None
        else:
            self._validator_type, option = validator
            self._validator = lambda text: validator_dict[self._validator_type](text, option)

    # ==================================================
    def set_read_only(self, flag):
        """
        Set read only.

        Args:
            flag (bool): read only ?
        """
        self._read_only = flag
        if flag:
            self.setStyleSheet(self._read_only_style)
        else:
            self.setStyleSheet(self._valid_style)

    # ==================================================
    def focusInEvent(self, event):
        """
        Focus-out event.
        """
        super().setText(self._raw)
        super().focusInEvent(event)
        self.selectAll()

    # ==================================================
    def focusOutEvent(self, event):
        """
        Focus-out event.
        """
        self.clearFocus()
        super().focusOutEvent(event)
        self.focusOut.emit()

    # ==================================================
    def clearFocus(self):
        """
        Clear focus.
        """
        if not self.read_only:
            if not self._valid:
                self.setStyleSheet(self._valid_style)
            self.setText(self._backup)
        super().clearFocus()

    # ==================================================
    def keyPressEvent(self, event):
        """
        Key event.

        Args:
            event (str): event.
        """
        k = event.key()
        if k == Qt.Key_Escape:
            self.clearFocus()
            self.setFocus()
        elif k == Qt.Key_Return:
            self.setText(self.text())
            self.close_edit()
        else:
            super().keyPressEvent(event)


# ==================================================
class HBar(QFrame):
    # ==================================================
    def __init__(self, parent=None):
        """
        Horizontal bar item.

        Args:
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)
        self.setMinimumSize(0, 10)
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)


# ==================================================
class Button(QPushButton):
    # ==================================================
    def __init__(self, parent=None, text="", toggle=False):
        """
        Button widget.

        Args:
            parent (QWidget, optional): parent.
            text (str, optional): text.
            toggle (bool, optional): toggle button ?
        """
        super().__init__(text, parent)
        self.setCheckable(toggle)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)


# ==================================================
class Combo(QComboBox):
    # ==================================================
    def __init__(self, parent=None, item=[], init=None):
        """
        Combo widget.

        Args:
            parent (QWidget, optional): parent.
            item (list, optional): list of items, [str].
            init (str, optional): initial value.
        """
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)
        self.set_item(item)
        if init is not None:
            self.setCurrentText(init)

    # ==================================================
    def get_item(self):
        """
        Get item.

        Returns:
            - (list) -- item list.
        """
        lst = [self.itemText(i) for i in range(self.count())]
        return lst

    # ==================================================
    def set_item(self, item):
        """
        Set item.

        Args:
            item (list): item list.
        """
        self.blockSignals(True)
        self.clear()
        self.addItems(item)
        self.blockSignals(False)

    # ==================================================
    def find_index(self, key):
        """
        Find index.

        Args:
            key (str): item key.

        Returns:
            - (int) -- index.
        """
        item = self.get_item()
        index = [idx for idx, s in enumerate(item) if key in s]
        return index


# ==================================================
class Spin(QSpinBox):
    # ==================================================
    def __init__(self, parent=None, min=0, max=1):
        """
        Spin widget.

        Args:
            parent (QWidget, optional): parent.
            min (int, optional): minimum value.
            max (int, optional): maximum value.
        """
        super().__init__(parent)
        self.setMinimum(min)
        self.setMaximum(max)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)


# ==================================================
class DSpin(QDoubleSpinBox):
    # ==================================================
    def __init__(self, parent=None, min=0.0, max=1.0, step=0.1):
        """
        Spin widget.

        Args:
            parent (QWidget, optional): parent.
            min (float, optional): minimum value.
            max (float, optioanl): maximum value.
            step (float, optional): step value.
        """
        super().__init__(parent)
        self.setMinimum(min)
        self.setMaximum(max)
        self.setSingleStep(step)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)


# ==================================================
class Check(QCheckBox):
    # ==================================================
    def __init__(self, parent=None, text=""):
        """
        Check widget.

        Args:
            parent (QWidget, optional): parent.
            text (str, optional): text.
        """
        super().__init__(text, parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)

    # ==================================================
    def is_checked(self):
        """
        Is checked ?

        Returns:
            - (bool) -- checked ?
        """
        return self.checkState() == Qt.Checked


# ==================================================
class VSpacer(QSpacerItem):
    # ==================================================
    def __init__(self):
        """
        Vertical spacer.
        """
        super().__init__(1, 1, QSizePolicy.Minimum, QSizePolicy.Expanding)


# ==================================================
class HSpacer(QSpacerItem):
    # ==================================================
    def __init__(self):
        """
        Horizontal spacer.
        """
        super().__init__(1, 1, QSizePolicy.Expanding, QSizePolicy.Minimum)


# ==================================================
class Editor(Panel):
    returnPressed = Signal(str)

    # ==================================================
    def __init__(self, parent=None, text="", validator=None, color="black", size=10, dpi=120):
        """
        Math equation label widget.

        Args:
            parent (QWidget, optional): parent.
            text (str, optional): text.
            validator (tuple, optional): (validator_type, option).
            color (str, optioanl): color name.
            size (int, optional): font size.
            dpi (int, optional): DPI.

        Note:
            - following validators can be used.
                - int: (min, max).
                - float: (min, max, digit).
                - sympy_float: (digit).
                - sympy: (variable list).
                - ilist: (shape).
                - list: (shape, variable list, digit).
                - site: (use variable?).
                - bond: (use variable?).
                - site_bond: (use variable?).
                - vector_site_bond: (use variable?).
                - orbital_site_bond: (use variable?).
        """
        super().__init__(parent)

        math = False
        if validator is not None and validator[0] == "sympy":
            math = True

        self._editor = LineEdit(None, text, validator)
        self._display = Label(None, self._editor.text(), color=color, size=size, math=math, dpi=dpi)

        self.layout.addWidget(self._display, 0, 0)
        self.layout.addWidget(self._editor, 0, 0)

        self._editor.hide()
        self._in_edit = False

        self._editor.returnPressed.connect(self.close_editor)
        self._editor.focusOut.connect(self.clearFocus)

    # ==================================================
    def close_editor(self):
        """
        Close editor.
        """
        self.clearFocus()
        if self._editor._valid:
            self.returnPressed.emit(self._editor.raw_text())

    # ==================================================
    def clearFocus(self):
        """
        Clear focus.
        """
        if self._editor._valid:
            self._display.setText(self._editor.text())
        self._editor.clearFocus()
        self._editor.hide()
        self._display.show()
        self._in_edit = False

    # ==================================================
    def mouseDoubleClickEvent(self, event):
        """
        Mouse double-click event.
        """
        if not self._in_edit and not self._editor.read_only:
            self._in_edit = True
            self._display.hide()
            self._editor.show()
            self._editor.setFocus()

    # ==================================================
    def mousePressEvent(self, event):
        """
        Mouse click event.
        """
        current = QApplication.focusWidget()
        if current:
            current.clearFocus()
        super().mousePressEvent(event)

    # ==================================================
    def text(self):
        """
        Text.

        Returns:
            - (str) -- text.
        """
        if self._editor._validator_type == "sympy":
            return self._editor.raw_text()
        else:
            return self._display.text()

    # ==================================================
    def setText(self, text):
        """
        Set text.

        Args:
            text (str): text.
        """
        self._editor.setText(text)
        if self._editor._valid:
            self._display.setText(self._editor.text())

    # ==================================================
    def raw_text(self):
        """
        Raw text.

        Returns:
            - (str) -- raw text.
        """
        return self._editor.raw_text()

    # ==================================================
    def setCurrentText(self, text):
        """
        Set current text.

        Args:
            text (str): text.
        """
        self._editor.setText(text)
        if self._editor._valid:
            self._display.setText(self._editor.text())

    # ==================================================
    def currentText(self):
        """
        Get current text.

        Returns:
            - (str) -- current text.
        """
        return self._editor.raw_text()


# ==================================================
class ColorSelector(QComboBox):
    # ==================================================
    def __init__(self, parent=None, current="", color_type="color"):
        """
        Color selector widget.

        Args:
            parent (QWidget, optional): parent.
            current (str, optional): default color.
            color_type (str, optional): color/colormap/color_both

        Notes:
            - connect currentTextChanged.
        """
        super().__init__(parent=parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(Qt.NoFocus)

        size = self.font().pointSize()

        color_pixmap, separator = color2pixmap(color_type, size)
        names = list(color_pixmap.keys())

        if current == "":
            current = names[0]
        try:
            current = names.index(current)
        except ValueError:
            current = 0

        self.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.blockSignals(True)
        for color, pixmap in color_pixmap.items():
            self.addItem(QIcon(pixmap), color, self)
        self.blockSignals(False)
        self.setCurrentIndex(current)

        size = next(iter(color_pixmap.values())).size()

        self.setIconSize(size)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.setFixedHeight(size.height() * 2.8)

        for i in separator:
            self.insertSeparator(i)



"""
About dialog.

This module provides about dialog for PyVistaWidget.
"""

from pathlib import Path
import sys
from numpy import __version__ as numpy_ver
from sympy import __version__ as sympy_ver
from matplotlib import __version__ as matplot_ver
from PySide6 import __version__ as pyside6_ver
import pyvista as pv
from pyvistaqt import __version__ as pyvistaqt_ver
from PySide6.QtWidgets import QDialog, QWidget, QDialogButtonBox, QSizePolicy
from PySide6.QtGui import QPixmap
from qtdraw.widget.custom_widget import Layout, Label, VSpacer
from qtdraw.__init__ import __version__, __date__, __author__
from qtdraw.util.util import check_multipie


# ==================================================
class AboutDialog(QDialog):
    # ==================================================
    def __init__(self, widget, parent=None):
        """
        About dialog.

        Args:
            widget (PyVistaWidget): widget.
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)
        self.setWindowTitle("About")
        self.resize(100, 100)
        policy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.setSizePolicy(policy)

        self.pvw = widget

        # buttons.
        button = QDialogButtonBox(QDialogButtonBox.Ok)
        button.accepted.connect(self.accept)

        panel = self.create_about_panel(self)

        # main layout
        layout = Layout(self)
        layout.setContentsMargins(20, 20, 20, 20)
        layout.addWidget(panel, 0, 0, 1, 1)
        layout.addWidget(button, 1, 0, 1, 1)

    # ==================================================
    def create_about_panel(self, parent):
        """
        Create about panel.
        """
        indent = " " * 4
        vtk_ver = ".".join(map(str, pv.vtk_version_info))
        pyvista_ver = pv._version.__version__
        python_ver = sys.version.replace(" [", f"\n{indent+indent}[")
        multipie = "version" in self.pvw._status["multipie"].keys()

        panel = QWidget(parent)
        layout = Layout(panel)

        label_icon = Label(parent)
        file = str(Path(__file__).parent / "qtdraw.png")
        pix = QPixmap(file)
        pix = pix.scaledToHeight(int(2.5 * self.height()))
        label_icon.setPixmap(pix)
        label_qtdraw = Label(parent, "QtDraw", True)
        label_copyright = Label(parent, self.pvw.copyright)
        label_python = Label(parent, f"\n{indent}Python: Ver. {python_ver}")
        label_pyvista = Label(parent, f"{indent}PyVista: Ver. {pyvista_ver}")
        label_pvqt = Label(parent, f"{indent}PyVistaQt: Ver. {pyvistaqt_ver}")
        label_vtk = Label(parent, f"{indent}VTK: Ver. {vtk_ver}")
        label_qt = Label(parent, f"{indent}PySide6: Ver. {pyside6_ver}")
        label_numpy = Label(parent, f"{indent}NumPy: Ver. {numpy_ver}")
        label_sympy = Label(parent, f"{indent}SymPy: Ver. {sympy_ver}")
        label_matplotlib = Label(parent, f"{indent}Matplotlib: Ver. {matplot_ver}")
        if multipie:
            multipie_ver = self.pvw._status["multipie"]["version"]
            label_multipie = Label(parent, f"{indent}MultiPie: Ver. {multipie_ver}")

        panel1 = QWidget(parent)
        layout1 = Layout(panel1)
        layout1.setContentsMargins(10, 10, 10, 10)
        layout1.addWidget(label_qtdraw, 0, 0, 1, 1)
        layout1.addWidget(label_copyright, 1, 0, 1, 1)
        layout1.addWidget(label_python, 2, 0, 1, 1)
        layout1.addWidget(label_pyvista, 3, 0, 1, 1)
        layout1.addWidget(label_pvqt, 4, 0, 1, 1)
        layout1.addWidget(label_vtk, 5, 0, 1, 1)
        layout1.addWidget(label_qt, 6, 0, 1, 1)
        layout1.addWidget(label_numpy, 7, 0, 1, 1)
        layout1.addWidget(label_sympy, 8, 0, 1, 1)
        layout1.addWidget(label_matplotlib, 9, 0, 1, 1)
        if multipie:
            layout1.addWidget(label_multipie, 10, 0, 1, 1)
            layout1.addItem(VSpacer(), 11, 0, 1, 1)
        else:
            layout1.addItem(VSpacer(), 10, 0, 1, 1)

        layout.addWidget(label_icon, 0, 0, 1, 1)
        layout.addWidget(panel1, 0, 1, 1, 1)

        return panel


# ==================================================
def get_version_info():
    """
    Get version info.

    Returns:
        - (str) -- version info.
    """
    indent = " " * 4
    vtk_ver = ".".join(map(str, pv.vtk_version_info))
    pyvista_ver = pv._version.__version__
    python_ver = sys.version
    cr = f"Versoin {__version__}, Copyright (C) {__date__} by {__author__}"

    s = "* QtDraw: " + cr + "\n"
    s += f"{indent}Python: Ver. {python_ver}" + "\n"
    s += f"{indent}PyVista: Ver. {pyvista_ver}" + "\n"
    s += f"{indent}PyVistaQt: Ver. {pyvistaqt_ver}" + "\n"
    s += f"{indent}VTK: Ver. {vtk_ver}" + "\n"
    s += f"{indent}PySide6: Ver. {pyside6_ver}" + "\n"
    s += f"{indent}NumPy: Ver. {numpy_ver}" + "\n"
    s += f"{indent}SymPy: Ver. {sympy_ver}" + "\n"
    s += f"{indent}Matplotlib: Ver. {matplot_ver}" + "\n"
    if check_multipie():
        from multipie import __version__ as multipie_ver

        s += f"{indent}MultiPie: Ver. {multipie_ver}" + "\n"
    s += "-" * 90

    return s



"""
Color scheme of elements for VESTA and Jmol.

This module contains color scheme of elements for VESTA and Jmol.
"""

# ==================================================
element_color = {
    "VESTA": {
        "Ac": "cornflowerblue",
        "Ag": "silver",
        "Al": "skyblue",
        "Am": "royalblue",
        "Ar": "gainsboro",
        "As": "yellowgreen",
        "At": "dimgrey",
        "Au": "gold",
        "B": "forestgreen",
        "Ba": "lime",
        "Be": "mediumaquamarine",
        "Bh": "crimson",
        "Bi": "darkorchid",
        "Bk": "mediumslateblue",
        "Br": "saddlebrown",
        "C": "darkslategrey",
        "Ca": "steelblue",
        "Cd": "khaki",
        "Ce": "lemonchiffon",
        "Cf": "darkorchid",
        "Cl": "lime",
        "Cm": "mediumslateblue",
        "Co": "mediumblue",
        "Cr": "darkblue",
        "Cs": "indigo",
        "Cu": "royalblue",
        "Db": "crimson",
        "Dy": "turquoise",
        "Er": "springgreen",
        "Es": "darkorchid",
        "Eu": "aquamarine",
        "F": "lightsteelblue",
        "Fe": "darkgoldenrod",
        "Fm": "darkorchid",
        "Fr": "indigo",
        "Ga": "lightgreen",
        "Gd": "turquoise",
        "Ge": "lightslategrey",
        "H": "pink",
        "He": "blanchedalmond",
        "Hf": "mediumturquoise",
        "Hg": "silver",
        "Ho": "mediumspringgreen",
        "Hs": "crimson",
        "I": "darkmagenta",
        "In": "grey",
        "Ir": "teal",
        "K": "blueviolet",
        "Kr": "mistyrose",
        "La": "limegreen",
        "Li": "lightgreen",
        "Lr": "mediumvioletred",
        "Lu": "forestgreen",
        "Md": "mediumvioletred",
        "Mg": "darkorange",
        "Mn": "darkmagenta",
        "Mo": "mediumaquamarine",
        "Mt": "crimson",
        "N": "lightsteelblue",
        "Na": "gold",
        "Nb": "mediumaquamarine",
        "Nd": "gainsboro",
        "Ne": "deeppink",
        "Ni": "silver",
        "No": "mediumvioletred",
        "Np": "dodgerblue",
        "O": "red",
        "Os": "teal",
        "P": "silver",
        "Pa": "deepskyblue",
        "Pb": "dimgrey",
        "Pd": "teal",
        "Pm": "aquamarine",
        "Po": "saddlebrown",
        "Pr": "lightgoldenrodyellow",
        "Pt": "lightgrey",
        "Pu": "dodgerblue",
        "Ra": "green",
        "Rb": "darkorchid",
        "Re": "steelblue",
        "Rf": "crimson",
        "Rh": "teal",
        "Rn": "steelblue",
        "Ru": "darkcyan",
        "S": "yellow",
        "Sb": "mediumpurple",
        "Sc": "mediumorchid",
        "Se": "chartreuse",
        "Sg": "crimson",
        "Si": "royalblue",
        "Sm": "aquamarine",
        "Sn": "darkgrey",
        "Sr": "lime",
        "Ta": "cornflowerblue",
        "Tb": "turquoise",
        "Tc": "lightseagreen",
        "Te": "darkgoldenrod",
        "Th": "deepskyblue",
        "Ti": "lightskyblue",
        "Tl": "sienna",
        "Tm": "limegreen",
        "U": "dodgerblue",
        "V": "red",
        "W": "dodgerblue",
        "Xe": "steelblue",
        "Y": "paleturquoise",
        "Yb": "limegreen",
        "Zn": "grey",
        "Zr": "lime",
    },
    "Jmol": {
        "Ac": "cornflowerblue",
        "Ag": "silver",
        "Al": "darkgrey",
        "Am": "royalblue",
        "Ar": "skyblue",
        "As": "orchid",
        "At": "dimgrey",
        "Au": "gold",
        "B": "lightpink",
        "Ba": "lime",
        "Be": "greenyellow",
        "Bh": "crimson",
        "Bi": "darkorchid",
        "Bk": "mediumslateblue",
        "Br": "brown",
        "C": "grey",
        "Ca": "lime",
        "Cd": "khaki",
        "Ce": "lemonchiffon",
        "Cf": "darkorchid",
        "Cl": "limegreen",
        "Cm": "mediumslateblue",
        "Co": "lightcoral",
        "Cr": "darkgrey",
        "Cs": "indigo",
        "Cu": "peru",
        "Db": "crimson",
        "Dy": "turquoise",
        "Er": "springgreen",
        "Es": "darkorchid",
        "Eu": "aquamarine",
        "F": "yellowgreen",
        "Fe": "chocolate",
        "Fm": "darkorchid",
        "Fr": "indigo",
        "Ga": "rosybrown",
        "Gd": "turquoise",
        "Ge": "slategrey",
        "H": "white",
        "He": "lightcyan",
        "Hf": "mediumturquoise",
        "Hg": "silver",
        "Ho": "mediumspringgreen",
        "Hs": "crimson",
        "I": "darkmagenta",
        "In": "grey",
        "Ir": "teal",
        "K": "darkorchid",
        "Kr": "mediumturquoise",
        "La": "lightskyblue",
        "Li": "violet",
        "Lr": "mediumvioletred",
        "Lu": "forestgreen",
        "Md": "mediumvioletred",
        "Mg": "chartreuse",
        "Mn": "mediumpurple",
        "Mo": "mediumaquamarine",
        "Mt": "crimson",
        "N": "royalblue",
        "Na": "mediumorchid",
        "Nb": "mediumaquamarine",
        "Nd": "gainsboro",
        "Ne": "paleturquoise",
        "Ni": "limegreen",
        "No": "mediumvioletred",
        "Np": "dodgerblue",
        "O": "red",
        "Os": "teal",
        "P": "darkorange",
        "Pa": "deepskyblue",
        "Pb": "dimgrey",
        "Pd": "teal",
        "Pm": "aquamarine",
        "Po": "saddlebrown",
        "Pr": "lightgoldenrodyellow",
        "Pt": "lightgrey",
        "Pu": "dodgerblue",
        "Ra": "green",
        "Rb": "darkorchid",
        "Re": "steelblue",
        "Rf": "crimson",
        "Rh": "teal",
        "Rn": "steelblue",
        "Ru": "darkcyan",
        "S": "yellow",
        "Sb": "mediumpurple",
        "Sc": "gainsboro",
        "Se": "orange",
        "Sg": "crimson",
        "Si": "wheat",
        "Sm": "aquamarine",
        "Sn": "slategrey",
        "Sr": "lime",
        "Ta": "cornflowerblue",
        "Tb": "turquoise",
        "Tc": "lightseagreen",
        "Te": "darkgoldenrod",
        "Th": "deepskyblue",
        "Ti": "silver",
        "Tl": "sienna",
        "Tm": "limegreen",
        "U": "dodgerblue",
        "V": "darkgrey",
        "W": "dodgerblue",
        "Xe": "steelblue",
        "Y": "paleturquoise",
        "Yb": "limegreen",
        "Zn": "lightslategrey",
        "Zr": "skyblue",
    },
}



"""
Versatile utility.

This module contains the utilities mainly for string and list.
"""

import numpy as np
from fractions import Fraction
from sympy.parsing.sympy_parser import parse_expr, standard_transformations, implicit_multiplication, rationalize
from sympy import SympifyError, Symbol, Basic


# ==================================================
def str_to_list(s):
    """
    Convert a string to a list of strings.

    Args:
        s (str): a string with irregular-shaped list format.

    Returns:
        - (list) -- a list of strings.

    Note:
        - in case of no bracket, return as it is.
        - raise ValueError for invalid brackets.
    """
    if s.count("[") != s.count("]"):
        raise ValueError(f"invalid blackets in '{s}'.")

    if s.count("[") == 0:
        return s

    nested_list = []
    stack = []
    current_word = ""

    for char in s:
        if char == "[":
            if current_word.strip():  # remove space, and append it for non-null string.
                stack[-1].append(current_word.strip())
                current_word = ""
            stack.append([])
        elif char == "]":
            if current_word.strip():  # remove space, and append it for non-null string.
                stack[-1].append(current_word.strip())
                current_word = ""
            if stack:
                popped = stack.pop()
                if stack:
                    stack[-1].append(popped)
                else:
                    nested_list.append(popped)
        elif char == ",":
            if current_word.strip():  # remove space, and append it for non-null string.
                stack[-1].append(current_word.strip())
                current_word = ""
        else:
            current_word += char

    if current_word.strip():  # parse last word.
        if stack:
            stack[-1].append(current_word.strip())
        else:
            nested_list.append(current_word.strip())

    return nested_list[0]


# ==================================================
def is_regular_list(lst):
    """
    Is regular-shaped list ?

    Args:
        lst (list): a list.

    Returns:
        - (bool) -- is regular-shaped list ?
    """
    try:
        np.array(lst)
        return True
    except ValueError:
        return False


# ==================================================
def apply_to_list(func, lst):
    """
    Apply a function to each element of a list.

    Args:
        func (function): a function.
        lst (list): a list.

    Returns:
        - (list) -- an applied list.

    Note:
        - applicable to an irregular-shaped list.
    """
    if not isinstance(lst, list):
        return func(lst)

    result = []
    for sub_lst in lst:
        if isinstance(sub_lst, list):
            # apply function to sub list recursively.
            result.append(apply_to_list(func, sub_lst))
        else:
            # apply function to non list.
            result.append(func(sub_lst))
    return result


# ==================================================
def apply_to_numpy(func, lst):
    """
    Apply a function to each element of ndarray.

    Args:
        func (function): a function.
        lst (ndarray): a numpy array.

    Returns:
        - (ndarray) -- an applied array.

    Note:
        - applicable only to a regular-shaped ndarray.
    """
    return np.vectorize(func)(lst)


# ==================================================
def to_fraction(x, max_denominator=1000000):
    """
    Convert a float number to a fractional one.

    Args:
        x (float): a float number.
        max_denominator (int, optional): max. of denominator.

    Returns:
        - (str) -- a fractional string.
    """
    return str(Fraction(x).limit_denominator(max_denominator))


# ==================================================
def get_variable(sp_ex):
    """
    Get variables used in a sympy expression.

    Args:
        sp_ex (sympy): a sympy expression (except for Matrix).

    Returns:
        - (list) -- variable strings (sorted).
    """
    var = set()

    if isinstance(sp_ex, Basic):
        var.update(set(map(str, sp_ex.atoms(Symbol))))

    return sorted(var)


# ==================================================
def str_to_sympy(s, check_var=None, rational=True, subs=None):
    """
    Convert a string to a sympy.

    Args:
        s (str): a string.
        local (dict, optional): variables to replace.
        check_var (list, optional): variables to accept.
        rational (bool, optional): use rational number ?
        subs (dict, optional): replace dict for local variables.

    Returns:
        - (sympy) -- a sympy.

    Notes:
        - if format error occurs, return None.
        - if s cannot be converted to a sympy, return None.
    """
    if check_var is None:
        check_var = []

    check_var = set(check_var)

    transformations = standard_transformations + (implicit_multiplication,)
    if rational:
        transformations += (rationalize,)

    try:
        expression = parse_expr(s, transformations=transformations, local_dict=subs)
    except (SympifyError, SyntaxError, TypeError):
        return None
    var = set(get_variable(expression))
    if len(check_var) != 0 and not (var <= check_var):
        return None

    return expression


# ==================================================
def str_to_numpy(s, digit=None, check_var=None, check_shape=None):
    """
    Convert a string (list) to a numpy array.

    Args:
        s (str): a string (list).
        digit (int, optional): accuracy digit (only for numerical vector).
        check_var (list, optional): variables to accept.
        check_shape (tuple, optional): shape to check.

    Returns:
        - (ndarray): a numpy array.

    Note:
        - when parse error occurrs, return None.
        - in check_shape, '0' means no check.
    """
    try:
        sl = str_to_list(s)
    except:
        return None
    if sl is None:
        return None
    sl = np.array(sl)
    if check_shape is not None:
        if sl.ndim != len(check_shape):
            return None
        shape = sl.shape
        for i in range(len(shape)):
            if check_shape[i] > 0 and shape[i] != check_shape[i]:
                return None

    try:
        is_ex = digit == None
        sl = apply_to_numpy(lambda x: str_to_sympy(x, check_var, rational=is_ex), sl)
        if np.any(sl == None):
            return None
        if not is_ex:
            sl = apply_to_numpy(float, sl).round(digit)
    except:
        return None

    return sl


# ==================================================
def affine_trans(v, s=None, A=None, digit=None, check_var=None):
    """
    Affine transformation, A.v + s.

    Args:
        v (str): site/vector or a list of site/vector.
        s (str, optional): shift vector or a list of shift vector.
        A (str, optional): rotaional matrix (3x3), [a1, a2, a3].
        digit (int, optional): accuracy digit (only for numerical vector).
        check_var (list, optional): variables to accept.

    Returns:
        - (ndarray) -- transformed vector.
    """
    v = str_to_numpy(v, digit, check_var)
    if v is None:
        return None
    if v.ndim != 1 and v.ndim != 2:
        return None
    if v.ndim == 1 and v.shape[0] != 3:
        return None
    if v.ndim == 2 and v.shape[1] != 3:
        return None

    if A is not None:
        A = str_to_numpy(A, digit, check_var, (3, 3))
        if A is None:
            return None

        v = v @ A.T

    if s is not None:
        s = str_to_numpy(s, digit, check_var)
        if s is None:
            return None
        if s.ndim != 1 and s.ndim != 2:
            return None
        if s.ndim == 1 and s.shape[0] != 3:
            return None
        if s.ndim == 2 and s.shape[1] != 3:
            return None

        if v.ndim == s.ndim:
            v += s
        else:
            if v.ndim == 1:
                v = np.tile(v, (s.shape[0], 1)) + s
            else:
                v = v + np.tile(s, (v.shape[0], 1))

    return v



"""
Logging utility.

This module provides utilities for logging.
"""

import sys
import os
import time
import logging
from functools import wraps
from PySide6.QtWidgets import QWidget, QPlainTextEdit, QGridLayout, QFileDialog
from PySide6.QtGui import QFont
from qtdraw.util.qt_event_util import ExceptionHook


# ==================================================
class LogHandler(logging.Handler):
    # ==================================================
    def __init__(self, level=logging.DEBUG, stream=None, text_widget=None):
        """
        Log handler.

        Args:
            level (Level, optional): log level.
            stream (TextIO, optional): stream.
            text_widget (LogWidget, optional): text widget.

        Note:
            - if stream/text_widget is None, it is not used.
        """
        super().__init__()

        # set handler.
        fmt = "%(asctime)s | %(levelname)8s | %(message)s"
        date_fmt = "%Y-%m-%d %H:%M:%S"
        self.setLevel(level)
        logging.basicConfig(level=level, format=fmt, datefmt=date_fmt, handlers=[self], encoding="utf-8")
        self.header = False

        # set plain text widget.
        if text_widget is not None:
            self.text_widget = text_widget

        # set stream.
        self.stream = stream

    # ==================================================
    def emit(self, record):
        """
        Emit message.

        Args:
            record (LogRecord): log record.
        """
        if not self.header:
            header = " Date Time          | Level    | Message"
            if self.text_widget is not None:
                self.text_widget.append_text(header)
            if self.stream is not None:
                print(header, file=self.stream)
            self.header = True

        msg = self.format(record)
        if self.text_widget is not None:
            self.text_widget.append_text(msg)
        if self.stream is not None:
            print(msg, file=self.stream)


# ==================================================
def timer(func):
    """
    Timer decorater.

    Args:
        func (Function): function to decorate.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        logging.info(f"=== ({func.__name__}) begin === ")
        result = func(*args, **kwargs)
        end = time.time()
        logging.info(f"=== ({func.__name__}) end ({end - start:.7f} [s] elapsed) ===")
        return result

    return wrapper


# ==================================================
def start_logging(level=logging.DEBUG, stream=None, text_widget=None, hook=True):
    """
    Start logging.

    Args:
        level (Level, optional): logging level.
        stream (TextIO, optional): stream for logging.
        text_widget (LogWidget, optional): text widget.
        hook (bool, optional): exception hook ?

    Note:
        - stream is True, sys.stderr is used.
    """
    if stream is True:
        stream = sys.stderr
    log_handler = LogHandler(level=level, stream=stream, text_widget=text_widget)
    logging.getLogger().addHandler(log_handler)
    if hook:
        ExceptionHook()


# ==================================================
class LogWidget(QWidget):
    # ==================================================
    def __init__(self, title="log message", level=logging.DEBUG, stream=None, hook=True, parent=None):
        """
        Log widget.

        Args:
            title (str, optional): window title.
            level (Level, optional): log level.
            stream (TextIO, optional): stream for logging.
            hook (bool, optional): exception hook ?
            parent (QObject, optional): parent.

        Note:
            - if stream is True, sys.stderr is used.
            - if level is None, no logging.
        """
        super().__init__(parent)
        self.log = QPlainTextEdit(self)
        font = QFont("Monaco", 11)
        font.setStyleHint(QFont.TypeWriter)
        self.log.setFont(font)
        self.log.setReadOnly(True)
        if level is not None:
            start_logging(level, stream, self, hook)

        self.setWindowTitle(title)
        self.resize(640, 800)

        layout = QGridLayout(self)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(10)
        layout.addWidget(self.log, 0, 0, 1, 1)

    # ==================================================
    def append_text(self, text):
        """
        Append text.

        Args:
            text (str): text.
        """
        self.log.appendPlainText(text)

    # ==================================================
    def set_text(self, text):
        """
        Set text.

        Args:
            text (str): text.
        """
        self.clear()
        self.append_text(text)

    # ==================================================
    def clear(self):
        """
        Clear log message.
        """
        self.log.clear()

    # ==================================================
    def save(self):
        """
        Save log message.
        """
        full = os.getcwd()
        filename, _ = QFileDialog.getSaveFileName(self, dir=full, filter="log (*.log)")
        if filename != "":
            with open(filename, "a") as f:
                print(self.log.plainText, file=f)



"""
Utility.

This module contains utilities for multipie plugin.
"""

import sympy as sp
from gcoreutils.nsarray import NSArray
from qtdraw.util.util_str import str_to_list


# ==================================================
def check_get_site(txt):
    """
    Check and get site.

    Args:
        txt (str): site.

    Returns:
        - (NSArray) -- site or None if invalid.
    """
    try:
        site = NSArray(txt)
        if site.style != "vector":
            return None
    except Exception:
        return None

    return site


# ==================================================
def check_get_bond(txt):
    """
    Check and get bond.

    Args:
        txt (str): bond.

    Returns:
        - (NSArray) -- bond or None if invalid.
    """
    try:
        bond = NSArray(txt)
        if bond.style not in ["bond", "bond_th", "bond_sv"]:
            return None
    except Exception:
        return None

    return bond


# ==================================================
def check_get_site_bond(txt, ret_site=False):
    """
    Check and get site (bond center) from site_bond.

    Args:
        txt (str): site_bond.
        ret_site (bool, optional): return site ?

    Returns:
        - (NSArray) -- site, bond or bond center, or None if invalid.
    """
    try:
        site_bond = NSArray(txt)
        if site_bond.style not in ["vector", "bond", "bond_th", "bond_sv"]:
            return None
    except Exception:
        return None

    if ret_site and site_bond.style != "vector":
        return site_bond.convert_bond("bond")[1]
    else:
        return site_bond


# ==================================================
def create_samb_object(z_samb, site, c_samb, z_head, irrep, pg, v, t_odd):
    """
    Create SAMB object.

    Args:
        z_samb (dict): combined SAMB.
        site (NSArray): site.
        c_samb (dict): cluster SAMB.
        z_head (str): multipole type.
        irrep (int): irrep. index.
        pg (PointGroup): point group.
        v (NSArray): vector variable.
        t_odd (bool): magnetic bond ?

    Returns:
        - (NSArray) -- (xyz)-polynomial at each cluster site.
    """
    obj = NSArray.zeros(len(site), "vector")
    for i in z_samb[z_head][irrep][1]:
        coeff, tag_h, tag_c = i
        harm = pg.harmonics[tag_h].expression(v=v)
        cluster = c_samb[tag_c]
        obj += coeff * harm * cluster

    if t_odd:
        obj *= -sp.I

    return obj


# ==================================================
def check_linear_combination(z_samb, form, head):
    """
    Check form of linear combination.

    Args:
        z_samb (dict): combined SAMB.
        form (str): expression of linear combination.
        head (str): multipole type.

    Returns:
        - (str) -- expression of linear combination (lower case).
        - (set) -- used variables.
        - (bool) -- magnetic bond ?
    """
    irrep_num = {i: len(z_samb[i]) for i in ["Q", "G", "T", "M"]}
    var_e = set([f"q{i+1:02d}" for i in range(irrep_num["Q"])] + [f"g{i+1:02d}" for i in range(irrep_num["G"])])
    var_m = set([f"t{i+1:02d}" for i in range(irrep_num["T"])] + [f"m{i+1:02d}" for i in range(irrep_num["M"])])

    form = form.lower()
    form_variable = set(NSArray(form).variable())

    t_odd = "Q"
    if form_variable.issubset(var_m):
        t_odd = "T"
    elif not form_variable.issubset(var_e):
        return None, None, None
    t_odd = head.replace("M", "T").replace("G", "Q") != t_odd

    return form, form_variable, t_odd


# ==================================================
def combined_format(tag_list):
    """
    Create formatted combined SAMB.

    Args:
        tag_list (tuple): (Z,X,Y) tag.

    Returns:
        - (str) -- formatted combined SAMB.
    """
    z_tag, x_tag, y_tag = tag_list
    t1 = (",".join(str(x_tag).split(",")[:-1]) + ")").replace("h", "a")
    t2 = ",".join(str(y_tag).split(",")[:-1]) + ")"
    tag = f"{z_tag} = {t1} x {t2}"

    return tag


# ==================================================
def parse_modulation_list(lst):
    """
    Parse modulation list.

    Args:
        lst (str): modulation list in str, [[basis,coeff,k,cos/sin]]

    Returns:
        - (list) -- modulation list.
    """
    try:
        lst = str_to_list(lst)
    except ValueError:
        return [[]]

    mod_data = []
    for i, row in enumerate(lst):
        basis, weight, k, cs = row
        mod_data.append([str(i), basis, weight, "[" + ",".join(k) + "]", cs])

    return mod_data



"""
Simple parser for VESTA.

This module contains the parser for VESTA.
"""

from pymatgen.core import Structure
from pymatgen.core import Lattice

# ==================================================
vesta_key = [  # VESTA Keywords.
    "#VESTA_FORMAT_VERSION",
    "ATOMM",
    "ATOMP",
    "ATOMS",
    "ATOMT",
    "BKGRC",
    "BONDM",
    "BONDP",
    "BONDS",
    "CELLP",
    "COMPS",
    "CONTR",
    "CRYSTAL",
    "DISPF",
    "DLATM",
    "DLBND",
    "DLPLY",
    "DPTHQ",
    "FORMM",
    "FORMP",
    "FORMS",
    "GROUP",
    "HBOND",
    "HKLPM",
    "HKLPP",
    "ISURF",
    "LABEL",
    "LBLAT",
    "LBLSP",
    "LIGHT0",
    "LIGHT1",
    "LIGHT2",
    "LIGHT3",
    "LMATRIX",
    "LORIENT",
    "LTRANSL",
    "MODEL",
    "PLN2D",
    "POLYM",
    "POLYP",
    "POLYS",
    "PROJT",
    "SBOND",
    "SCENE",
    "SECCL",
    "SECTP",
    "SECTS",
    "SHAPE",
    "SITET",
    "SPLAN",
    "STRUC",
    "STYLE",
    "SURFM",
    "SURFS",
    "SYMOP",
    "TEX3P",
    "TEXCL",
    "THERI",
    "TITLE",
    "UCOLP",
    "VECTR",
    "VECTS",
    "VECTT",
]


# ==================================================
def parse_vesta(filename):
    """
    parse VESTA file.

    Args:
        filename (str): filename.

    Returns:
        - (dict) -- VESTA keyword - content dict.
    """
    data = []
    with open(filename) as f:
        for line in f:
            d = line.split(" ")
            d = [i.replace("\n", "") for i in d if i.replace("\n", "") != ""]
            if d:
                data += [d + ["\n"]]
    data = sum(data, [])
    pos = sorted([(data.index(i), i) if i in data else (-1, i) for i in vesta_key], key=lambda x: x[0])

    dic = {}
    for i in range(len(pos) - 1):
        if pos[i][0] != -1:
            if pos[i + 1][0] - pos[i][0] != 1:
                dic[pos[i][1]] = (" ".join(data[pos[i][0] + 1 : pos[i + 1][0]])).strip()
            else:
                dic[pos[i][1]] = ""

    return dic


# ==================================================
def create_structure_vesta(vesta_dict):
    """
    create Structure from vesta dict.

    Args:
        vesta_dict (dict): vesta dict.

    Returns:
        - (Structure) -- structure.
    """
    space_group = int(vesta_dict["GROUP"].split(" ")[0])
    lattice = [float(i) for i in vesta_dict["CELLP"].split("\n")[0].split(" ") if i != ""]
    sites = [[j for j in i.split(" ") if j != ""] for i in vesta_dict["STRUC"].split("\n")[:-1:2]]
    species = [i[1] for i in sites]
    coords = [[float(i[4]), float(i[5]), float(i[6])] for i in sites]

    lattice = Lattice.from_parameters(*lattice)
    structure = Structure.from_spacegroup(space_group, lattice, species, coords)

    return structure



"""
Color selector widget.

This module provides color selector widget.
"""

import numpy as np
from matplotlib import cm
from PySide6.QtGui import QPixmap, QColor, QImage
from PySide6.QtCore import Qt
from gcoreutils.basic_util import apply
from gcoreutils.color_palette import all_colors, all_colormaps, check_color, name_sep


# ==================================================
def _colormap2pixmap(colormap, step=50):
    """
    Convert colormap to pixmap.

    Args:
        colormap (str): colormap.
        step (int, optional): division of colorbar gradation.

    Returns:
        - (QPixmap) -- color pixmap.
    """
    if colormap not in all_colormaps:
        raise ValueError(f"unknown colormap, {colormap} is given.")

    sm = cm.ScalarMappable(cmap=colormap.strip("*"))
    sm.norm.vmin = 0.0
    sm.norm.vmax = 1.0
    inds = np.linspace(0, 1, step)
    rgbas = sm.to_rgba(inds)
    rgbas = [QColor(int(r * 255), int(g * 255), int(b * 255), int(a * 255)).rgba() for r, g, b, a in rgbas]
    im = QImage(step, 1, QImage.Format_Indexed8)
    im.setColorTable(rgbas)
    for i in range(step):
        im.setPixel(i, 0, i)
    im = im.scaled(100, 100)
    pm = QPixmap.fromImage(im)
    return pm


# ==================================================
def _color2pixmap(color, color_type, size, aspect_ratio=4, step=50):
    """
    Convert color/colormap to pixmap.

    Args:
        color (str): color or colormap.
        color_type (str): color type, "color/colormap/color_both".
        size (int): vertical size (pixel).
        aspect_ratio (int, optional): aspect ratio of pximap (width/height).
        step (int, optional): division of colorbar gradation.

    Returns:
        - (QPixmap) -- color pixmap.
    """
    if check_color(color):
        if color_type == "color":
            colorbox = QPixmap(size, size)
        else:
            colorbox = QPixmap(aspect_ratio * size, size)
        colorbox.fill(QColor(all_colors[color][0]))
    else:
        colorbox = _colormap2pixmap(color, step=step)
        colorbox = colorbox.scaled(aspect_ratio * size, size, Qt.IgnoreAspectRatio, Qt.SmoothTransformation)
    return colorbox


# ==================================================
def color2pixmap(color_type, size, aspect_ratio=4, step=50):
    """
    Convert color/colormap to QPixmap.

    Args:
        color_type (str): color type, "color/colormap/color_both".
        size (int): size of pixmap.
        aspect_ratio (int, optional): aspect ratio of pximap (width/height).
        steps (int, optional): division of colorbar gradation.

    Retruns:
        - (dict) -- pixmap dict {name: pixmap}.
        - (list) -- separator position.
    """
    names, sep = name_sep(color_type)

    pixmap = apply(lambda c: _color2pixmap(c, color_type, size, aspect_ratio, step), names)
    pixmap_dict = {name: value for name, value in zip(names, pixmap)}

    return pixmap_dict, sep



"""
Simple parser for CIF, XSF, and VESTA.

This module contains the parser.
"""

from qtdraw.parser.util import parse_material, draw_site_bond


# ==================================================
def read_draw(filename, widget):
    """
    Read and draw CIF, XSF, VESTA file.

    Args:
        filename (str): filename.
        widget (PyVistaWidget): PyVista widget.

    Returns:
        - (dict) -- all data.
    """
    all_data, site_info, bond_info, symmetrized = parse_material(filename)
    widget.write_info("* " + str(symmetrized))

    name = all_data["status"]["model"]
    draw_site_bond(widget, name, site_info, bond_info)

    if filename.endswith(".xsf"):
        pos = site_info[0][3]
        widget.add_isosurface(data=filename, position=pos, color="Pastel1", surface="phase")

    return all_data



"""
TableView widget.

This module provides table view widget.
"""

from PySide6.QtWidgets import QTableWidget
from qtdraw.widget.custom_widget import Label


# ==================================================
class TableView(QTableWidget):
    # ==================================================
    def __init__(self, parent=None, data=[[""]], header=None, vertical=False, color="black", size=12, dpi=120):
        """
        Table view (math).

        Args:
            parent (QWidget, optional): parent.
            data (list, optional): table data in LaTeX code without "$".
            header (list, optional): header string.
            vertical (bool, optional): show vertical number header ?
            color (str, optional): color name.
            size (int, optional): font size.
            dpi (int, optional): DPI.
        """
        super().__init__(parent)

        row = len(data)
        column = len(data[0])

        self.setRowCount(row)
        self.setColumnCount(column)
        if header is not None:
            self.setHorizontalHeaderLabels(header)
            self.horizontalHeader().setStyleSheet("font-weight: bold;")
        else:
            self.horizontalHeader().setVisible(False)
        self.verticalHeader().setVisible(vertical)

        for i, r in enumerate(data):
            for j, item in enumerate(r):
                item = str(item)
                if item != "":
                    label = Label(self, item, color=color, size=size, math=True, dpi=dpi)
                    label.setContentsMargins(5, 5, 5, 5)
                    self.setCellWidget(i, j, label)

        self.resizeColumnsToContents()
        self.resizeRowsToContents()
